/**
 *   Copyright 2008 Marvin Herman Froeder
 * -->
 * <!--
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * -->
 *
 * <!--
 *     http://www.apache.org/licenses/LICENSE-2.0
 * -->
 *
 * <!--
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.sonatype.flexmojos.utilities;

import java.io.File;
import java.io.FileFilter;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

public class SourceFileResolver
{

    /**
     * Resolve a source file in the source paths
     * 
     * @param project maven project
     * @param sourceFileName sugested name on pom
     * @return source file or null if source not found
     */
    public static File resolveSourceFile( List<String> sourcePaths, String sourceFileName, String groupId,
                                          String artifactId )
    {
        List<File> sourceRoots = new ArrayList<File>();
        for ( String sourcePath : sourcePaths )
        {
            File sourceRoot = new File( sourcePath );
            if ( sourceRoot.isDirectory() )
            {
                sourceRoots.add( sourceRoot );
            }
        }

        return resolveSourceFile( sourceFileName, sourceRoots, groupId, artifactId );
    }

    /**
     * Resolve a source file in the source paths
     * 
     * @param project maven project
     * @param sourceFileName sugested name on pom
     * @return source file or null if source not found
     */
    public static File resolveSourceFile( String sourceFileName, List<File> sourceRoots, String groupId,
                                          String artifactId )
    {

        if ( sourceFileName != null )
        {
            File absoluteSourceFile = new File( sourceFileName );
            if ( absoluteSourceFile.isAbsolute() )
            {
                return absoluteSourceFile;
            }

            for ( File sourceDirectory : sourceRoots )
            {
                File sourceFile = new File( sourceDirectory, sourceFileName );
                if ( !sourceFile.exists() )
                {
                    continue;
                }
                return sourceFile;
            }

            return null;
        }

        for ( File sourceDirectory : sourceRoots )
        {
            File sourceFile = resolveFile( sourceDirectory, artifactId );
            if ( sourceFile == null )
            {
                sourceFile = resolveSourceFileByGroupIdAndArtifactId( sourceDirectory, groupId, artifactId );
            }

            if ( sourceFile != null )
            {
                return sourceFile;
            }
        }
        return null;
    }

    private static File resolveFile( File sourceDirectory, String artifactId )
    {
        if ( !sourceDirectory.isDirectory() )
        {
            return null;
        }

        File[] files = sourceDirectory.listFiles( new FileFilter()
        {
            public boolean accept( File pathname )
            {
                return pathname.isFile()
                    && ( pathname.getName().endsWith( ".mxml" ) || pathname.getName().endsWith( ".as" ) || pathname.getName().endsWith(
                                                                                                                                        ".css" ) );
            }
        } );

        if ( files.length == 1 )
        {
            return files[0];
        }
        if ( files.length > 1 )
        {
            for ( File file : files )
            {
                if ( file.getName().equalsIgnoreCase( "Main.mxml" ) || file.getName().equalsIgnoreCase( "Main.as" ) )
                {
                    return file;
                }
            }

            for ( File file : files )
            {
                if ( file.getName().equalsIgnoreCase( "Index.mxml" ) || file.getName().equalsIgnoreCase( "Index.as" ) )
                {
                    return file;
                }
            }

            if ( artifactId != null )
            {
                for ( File file : files )
                {
                    if ( file.getName().equalsIgnoreCase( artifactId + ".mxml" )
                        || file.getName().equalsIgnoreCase( artifactId + ".as" ) )
                    {
                        return file;
                    }
                }
            }

            List<File> appFiles = new ArrayList<File>();
            for ( File file : files )
            {
                if ( file.getName().endsWith( ".mxml" ) && isApplicationFile( file ) )
                {
                    appFiles.add( file );
                }
            }

            if ( appFiles.size() == 1 )
            {
                File file = appFiles.get( 0 );
                return file;
            }

        }

        return null;
    }

    private static File resolveSourceFileByGroupIdAndArtifactId( File sourceDirectory, String groupId, String artifactId )
    {
        if ( groupId == null )
        {
            return null;
        }

        String localPath = groupId.replace( '.', File.separatorChar );

        File packageDirectory = new File( sourceDirectory, localPath );
        if ( !packageDirectory.isDirectory() )
        {
            // there is no dir, so no file =D
            return null;
        }

        // let's try just groupId
        File sourceFile = resolveFile( packageDirectory, artifactId );
        if ( sourceFile != null )
        {
            return sourceFile;
        }

        if ( groupId.endsWith( artifactId ) )
        {
            return null;
        }

        packageDirectory = new File( packageDirectory, artifactId );

        sourceFile = resolveFile( packageDirectory, artifactId );

        return sourceFile;
    }

    /**
     * Parse an MXML file and returns true if the file is an application one
     * 
     * @param file the file to be parsed
     * @return true if the file is an application one
     */
    private static boolean isApplicationFile( File file )
    {
        try
        {
            SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
            parser.getXMLReader().setFeature( "http://xml.org/sax/features/namespaces", true );
            parser.getXMLReader().setFeature( "http://xml.org/sax/features/namespace-prefixes", true );
            ApplicationHandler h = new ApplicationHandler();
            parser.parse( file, h );
            return h.isApplicationFile();
        }
        catch ( Exception e )
        {
            return false;
        }
    }
}
